/**
  \file minezj.java
  Java applet minesweeper game implementation.
  */

import java.applet.Applet;
import java.util.Random;
import java.awt.*;

/** The java.applet.Applet extending class which represents simple implementation of minesweeper game. */
public class minezj extends java.applet.Applet
{
	private Image buffer;		///< Screen buffer.
	private Graphics screen;	///< Screen destination.

	private byte[] pground;		///< Battlefield.
	private Image[] img;		///< Array containing all sprites.
	private boolean fl_done;	///< Game state (true => game done).
	private int screen_x;		///< Applet screen width.
	private int screen_y;		///< Applet screen height.
	private int bgcolor = 0xE0E0E0;	///< Default background color.
	private int nr_mines = 85;	///< Number of mines in the game.
	private int size_x = 30;	///< Playground's width (number of cells).
	private int size_y = 15;	///< Playground's height (number of cells).
	private int nr_uncovered = 0;	///< Count of uncovered fields.

	private final byte F_MINE = 9;		///< Mine field.
	private final byte B_UNCOVERED = 0x10;  ///< Uncovered field.
	private final byte B_MARKED = 0x20;	///< Marked field (mine suspection).
	private final byte COUNT_MASK = 0x0F;	///< Low-nibble mask. It holds count of surrounding mines (0-8).
	private final byte C_SIZE = 15;	///< Cell size in pixels.

/** Provides applet information text. */
	public String getAppletInfo ()
	{
		return "minezj 1.0\r\nOndra Havel <xohavel@fd.cvut.cz>\r\n";
	}
	
/** Applet initialization routine. Acquires basic information and loads graphics from disk. */
	public void init ()
	{
		String opt;

		screen_x = getSize().width;
		screen_y = getSize().height;

		try {
			buffer = createImage (screen_x, screen_y);
			screen = buffer.getGraphics ();
		} catch (Exception e) {
			screen = null;
		}

		pground = new byte[size_x * size_y];

		opt = getParameter("bgcolor");
		if(opt!=null)
			bgcolor = Integer.parseInt(opt);
			
		img = new Image[13];
		for (int i = 0; i < 13; i++)
			img[i] = getImage (getCodeBase (), "./gfx/" + (i) + ".gif");
	}

/** Prepares all stuff for new game. Initializes needed variables, and generates
new playground. */
	public void reset_game ()
	{
		Random a;
		int todo, fields, mns = nr_mines, i, j, r, cnt;

		fl_done = false;
		nr_uncovered = 0;
		
		a = new Random ();
		todo = fields = size_y * size_x;

/*  clear the mine-field */
		for (i = 0; i < todo; i++)
			pground[i] = 0;

		while(mns>0) {
			mns--;
			r = (int)(todo * a.nextDouble());
			i = 0;
			cnt = 0;
			while (i<fields) {
				if(pground[i]==0) {
					if(cnt == r) {
						pground[i] = F_MINE;
						break;
					}
					cnt++;
				}	
				i++;
			}
		}

/*  count values for non-mine fields     */
		for(i=0;i<fields;i++) {
			if(pground[i]==0) {
				int y=i/size_x, x=i % size_x;
				/*  up */
				if(y!=0) {
					if(pground[i-size_x] == F_MINE)
						pground[i]++;
					/*  and left */
					if(x-1>=0 && pground[i - size_x - 1] == F_MINE)
						pground[i]++;
					/*  and right */
					if(x+1<size_x && pground[i-size_x+1] == F_MINE)
						pground[i]++;	
				}

			/* down */
			if(y!=size_y-1) {
				if(pground[i+size_x]==F_MINE)
					pground[i]++;
				/* and left */
				if(x-1>=0 && pground[i+size_x-1] == F_MINE)
					pground[i]++;
				/* and right */
				if(x+1<size_x && pground[i+size_x+1]==F_MINE)
					pground[i]++;
				}
				
				/* left */
				if(x!=0 && pground[i-1]==F_MINE)
					pground[i]++;
					
					/* right */
					if(x!=size_x-1&&pground[i+1]==F_MINE)
						pground[i]++;
			}
		}
	}

	/** Starts a brand new game. */
	public void start ()
	{
		reset_game ();
		repaint ();
	}

	/** Queries is empty and there are no mine fields surrounding it.
	  \param x Field x-coordinate.
	  \param y Field y-coordinate.
	  \returns boolean True if the field specified by [x,y] is empty and there's no mine field in it's neighbourhood. */
	private boolean is_empty(int x, int y)
	{
		return(pground[x + y * size_x] == 0);
	}

	/** Uncovers a field specified by position.
	  \param x Field x-coordinate.
	  \param y Field y-coordinate.
	  */
	private void revealcount(int x, int y)
	{
		if((pground[x+y*size_x] & B_UNCOVERED)==0 && (pground[x+y*size_x] & COUNT_MASK)!=0) {
			nr_uncovered++;
			pground[x+y*size_x] |= B_UNCOVERED;
		}
	}

	/** Recursively uncover all adjoining empty fields starting at specified position. This method may be called recursively.
	  \param x Starting field x-coordinate.
	  \param y Starting field y-coordinate.
	  */
	private void zeroafter909(int x, int y)
	{
		pground[x+y*size_x] = B_UNCOVERED;
		nr_uncovered++;
		if(x>0) {
			/*  left */
			if(is_empty(x - 1,y)) {
				zeroafter909(x - 1,y);
			} else {
				revealcount(x-1,y);
				if(y>0)
					revealcount(x-1,y-1);
				if(y<size_y-1)	
					revealcount(x-1,y+1);
			}
		}
		if(x<size_x-1) {
			/*  right */
			if(is_empty(x+1,y)) {
				zeroafter909(x+1,y);
			} else {
				revealcount(x+1,y);
				if(y>0)
					revealcount(x+1,y-1);
				if(y<size_y-1)	
					revealcount(x+1,y+1);
			}
		}
		if(y>0)	{
			/*  up */
			if(is_empty(x,y-1)) {
				zeroafter909(x,y-1);
			} else {
				revealcount(x,y-1);
			}
		}
		if(y<size_y-1)	{
			/* down */
			if(is_empty(x,y+1)) {
				zeroafter909(x,y+1);
			} else {
				revealcount(x,y+1);
			}
		}
	}

	/** Evaluate mouse button press event and perform action according to current state of the game. In case that game is not running new one is started. In case that game runs left click reveals a field and right click toggles mine suspection flag.
	  \param e Event description object.
	  \param x Mouse x-coordinate.
	  \param y Mouse y-coordinate.
	  */
	public boolean mouseDown (Event e, int x, int y)
	{
		int cell_x = x / C_SIZE, cell_y = y / C_SIZE;
		int off = cell_x + cell_y * size_x;
		
/*  start new game if old already ended */
		if (fl_done) {
			reset_game ();
			repaint ();
			return (true);
		}

/*  tide is high */
		if(cell_x > size_x || cell_y > size_y)
			return (true);

/*  uncovered field, do nothing */
		if((pground[off] & B_UNCOVERED)!=0)
			return (true);

/*  right click? tag/mark/untag/unmark it... */
		if ((e.modifiers & Event.META_MASK) == Event.META_MASK) {
			pground[off] ^= B_MARKED;	/*  toggle the mark bit */
			repaint();
			return (true);
		}

/*  Covered empty field, look around. Here may lay surrounding mines */
/*  which ought to be 'automatically' uncovered, zeroafter909 does */
/*  this job. */
		if(pground[off]==0) {
			zeroafter909(cell_x, cell_y);
			repaint();
			return (true);
		}

/*  do uncover if everything else failed...       */
		if((pground[off] & COUNT_MASK)==F_MINE) {
			fl_done = true;
		} else {
			pground[off] |= B_UNCOVERED;
			nr_uncovered++;
			if (nr_mines == size_x * size_y - nr_uncovered)
				fl_done = true;
		}

		repaint ();

		return (true);
	}

	/** Paint applet's window.
	  \param g Graphics object to paint to.
	 */
	public void paint (Graphics g)
	{
		update (g);
	}

	/** update applet's window.
	  \param g Graphics object to paint to.
	 */
	public synchronized void update (Graphics g)
	{
		if (screen != null) {
			do_draw (screen);
			g.drawImage (buffer, 0, 0, this);
		} else {
			do_draw (g);
		}
	}

	/** Repaint whole scene.
	  \param g Graphics object to paint to.
	 */
	private void do_draw (Graphics g)
	{
		int i, pic_nr, val, count;

		g.setColor (new Color (bgcolor));
		g.fillRect (0, 0, screen_x, screen_y);

		for (i = 0; i < size_x * size_y; i++) {
			val = pground[i];
			count = val & COUNT_MASK;

/*  check whether you are already dead or not */
			if((val & B_UNCOVERED)!=0 && count==F_MINE)
					fl_done = true;

/*  already done? well, unhide everything and show mistakes */
			if (fl_done) {
				if ((val & B_MARKED)!=0) { /* marked? */
					if(count!=F_MINE) {	/*  invalid mark */
						pic_nr = 12;
					} else {
						pic_nr = 11;	/*  common mark */
					}
				} else {
					pic_nr = count;
				}
			} else {	/*  still staying alive */
				if((val & B_UNCOVERED) == 0) {	/*  covered? :] */
					if((val & B_MARKED)!=0)	{	/*  marked? */
						pic_nr = 11;
					} else {
						pic_nr = 10;
					}
				} else {
					pic_nr = count;
				}
			}

			if (img[pic_nr] != null)
				g.drawImage (img[pic_nr], (i % size_x) * C_SIZE,
				(i / size_x) * C_SIZE, this);
		}

/*     g.setFont (getFont()); */
		g.setColor (Color.black);
		g.drawString ("nr_uncovered=" + nr_uncovered, 0, size_y * C_SIZE + 16);
		
		if(fl_done) {
			if (nr_mines == size_x * size_y - nr_uncovered) {
		g.drawString ("You have won!", (size_x * C_SIZE) >> 1, size_y * C_SIZE + 16);
			} else {
			g.drawString ("You have lost the match!", (size_x * C_SIZE) >> 1, size_y * C_SIZE + 16);
			}
		}
	}
}

/**
\mainpage MinezJ


\section s0 Introduction

Hello, this is MinezJ, Java applet implementation of famous minesweeper game. Source code and generated documentation can be found here, full source distribution is available <a href="../minezj-1.0.zip">here</a>.

\htmlonly
<p align=center>
<applet code="minezj.class" codebase="../" width=450 height=250>
<param name=bgcolor value=16777215>
</applet>
</p>
\endhtmlonly


\section s1 Compilation

Compile with "javac -target 1.1" in order to make it working with Microsoft Internet Explorer with Microsoft Java Virtual Machine. It won't work otherwise returning "applet can't be instantiated" (use of inappropriate or JDK deprecated methods) or "class minezj not found" could appear.


\section s2 Playground Representation

Playground is represented by byte array, one byte for one cell. Byte's low nibble (4 least significant bits) contains count of surrounding mine fields (0-8). If there's a mine on a field, it's low nibble's value is 9.

<ul>
<li>Bit 4 (0x10) indicates whether the field is revealed (=1).
<li>Bit 5 (0x20) indicates whether the field is marked (=1).
</ul>

\section s3 Applet Invocation

The applet can be invocated via applet tag where parameter "bgcolor" may be passed which specifies the background color of the applet. Note that the value has to be specified in decadic format. It is standard RGB color specification, one byte per color. In this case the 16777215 = 0xFFFFFF which is white color. If the applet resides in other directory the codebase has to be set accordingly.

<code>
<applet code="minezj.class" codebase="../" width=450 height=250><br>
<param name=bgcolor value=16777215><br>
</applet><br>
</code>

Have fun! Ondra Havel, 2006-1-27, Darmstadt, Germany
*/
